
using System;
using System.Collections.Generic;

namespace Animrig
{
    public class SpineRenderer : StateRenderer
    {
        // property
        protected Dictionary<string, Spine.Bone> BoneCacheList = new Dictionary<string, Spine.Bone>();
        public static readonly int SaveVersion = 1; 
        public Spine.AnimationState state = null;
        public Spine.Skeleton skeleton = null;
        public float AntiScaling = 1.0f;    // anti scaling. on unity, data scale x0.01.
        // public
        public SpineRenderer()
            : base()
        {
            ObjectType = ObjectType.Spine;

            Clear();
        }
        // common 
        public override void Clear()
        {
            base.Clear();

            state = null;
            skeleton = null;
            BoneCacheList.Clear();
        }
        public override void Update(double deltatime, double deltaphysics)
        {
            if (state == null) return;
            if (skeleton == null) return;

            state.Update((float)deltatime);
            state.Apply(skeleton);
            skeleton.UpdateWorldTransform();
        }
        // spine
        public override void SetAnimation(int trackindex, int usetrack, string basename, bool loop, bool addanim, MixBlend trackmix, float duration, bool seekmode, float loopbegin, float loopend)
        {
            void inSetEmpty(int index, float dt)
            {
                if (dt < 0.0) dt = state.Data.DefaultMix;

                if (addanim)
                    state.AddEmptyAnimation(index, dt, 0.0f);
                else
                    state.SetEmptyAnimation(index, dt);
            }
            void inSetAnimation(int index, string name1, string name2, float dt)
            {
                // search
                var anim = state.Data.SkeletonData.FindAnimation(name1);
                if (anim == null) anim = state.Data.SkeletonData.FindAnimation(name2);

                if (anim == null)
                {
                    inSetEmpty(index , dt);
                }
                else
                {
                    // set animation
                    Spine.TrackEntry track;
                    if (addanim)
                        track = state.AddAnimation(index, anim, loop, 0.0f);
                    else
                        track = state.SetAnimation(index, anim, loop);
                    if (track != null)
                    {
                        // mix blend
                        if ((track.MixBlend == Spine.MixBlend.Replace)||(track.MixBlend == Spine.MixBlend.Add))     // safe code
                        {
                            switch (trackmix)
                            {
                                case MixBlend.Add:
                                    track.MixBlend = Spine.MixBlend.Add;
                                    break;
                                case MixBlend.Replace:
                                    track.MixBlend = Spine.MixBlend.Replace;
                                    break;
                            }
                        }
                        // duration
                        if (dt < 0.0f)
                        {
                            // use default duration
                        }
                        else
                        {
                            track.MixDuration = Math.Max(dt, 0.0f);
                        }
                        // seek mode
                        if (seekmode)
                            track.TimeScale = 0.0f;
                        else
                            track.TimeScale = 1.0f;
                        // loop area
                        if (loopbegin != loopend)
                        {
                            track.AnimationStart = loopbegin;
                            track.AnimationEnd = loopend;
                            // it often makes sense to set TrackEntry.Animation Last to the same value to prevent timeline keys before the start time from triggering.
                            track.AnimationLast = loopbegin;
                        }
                    }
                }
            }

            if (state == null) return;
            
            // layertype AnimBlend1
            if (usetrack == 1)
            {
                if (basename == "empty")
                {
                    inSetEmpty(trackindex + 0, duration);
                }
                else
                {
                    inSetAnimation(trackindex + 0, basename, basename + "@up", duration);
                }
                return;
            }
            
            // layertype AnimBlend4
            if (usetrack == 4)
            {
                if (basename == "empty")
                {
                    inSetEmpty(trackindex + 0, duration);
                    inSetEmpty(trackindex + 1, duration);
                    inSetEmpty(trackindex + 2, duration);
                    inSetEmpty(trackindex + 3, duration);
                }
                else
                {
                    inSetAnimation(trackindex + 0, basename, basename + "@up", duration);
                    inSetAnimation(trackindex + 1, basename + "@right", "", duration);
                    inSetAnimation(trackindex + 2, basename + "@down", "", duration);
                    inSetAnimation(trackindex + 3, basename + "@left", "", duration);
                }
            }

            // 2 animations blend layer. use @up , @right.
            if (usetrack == 2)
            {
                if (basename == "empty")
                {
                    inSetEmpty(trackindex + 0, duration);
                    inSetEmpty(trackindex + 1, duration);
                }
                else
                {
                    inSetAnimation(trackindex + 0, basename, basename + "@up", duration);
                    inSetAnimation(trackindex + 1, basename + "@right", "", duration);
                }
            }
        }
        public override void SetTrackBlendAlpha(int trackindex, int usetrack, float alpha1, float alpha2, float alpha3, float alpha4, bool exists)
        {
            void inSet(int index, float alpha)
            {
                var track = state.GetCurrent(index);
                if (track == null) return;

                track.Alpha = alpha;
            }

            if (state == null) return;

            if (usetrack == 1)
            {
                inSet(trackindex + 0, alpha1);
            }
            
            if (usetrack == 4)
            { 
                inSet(trackindex + 0, alpha1);
                inSet(trackindex + 1, alpha2);
                inSet(trackindex + 2, alpha3);
                inSet(trackindex + 3, alpha4);
            }

            if (usetrack == 2)
            {
                // @right , @down are invert animation
                inSet(trackindex + 0, alpha1 - alpha3);
                inSet(trackindex + 1, alpha2 - alpha4);
            }
        }
        public override void SetTrackSeekTime(int trackindex, int usetrack, float time1, float time2, float time3, float time4, bool exists)
        {
            void inSet(int index, float time)
            {
                var track = state.GetCurrent(index);
                if (track == null) return;

                track.TrackTime = time;
            }

            if (state == null) return;

            if (usetrack == 1)
            {
                inSet(trackindex + 0, time1);
            }
            else
            {
                inSet(trackindex + 0, time1);
                inSet(trackindex + 1, time2);
                inSet(trackindex + 2, time3);
                inSet(trackindex + 3, time4);
            }
        }
        public override void SetMixDuration(float duration)
        {
            if (state == null) return;

            state.Data.DefaultMix = duration;
        }
        public override void SetSkin(string skinname)
        {
            if (skeleton == null) return;

            var skin = skeleton.Data.FindSkin(skinname);
            if (skin != null)
            {
                skeleton.SetSkin(skin);
                skeleton.SetToSetupPose();
            }
        }
        public override Matrix4 GetBoneMatrix(string bonename)
        {
            if (skeleton == null) return Misc.IdentityMatrix4;
            if (state == null) return Misc.IdentityMatrix4;

            // cache
            Spine.Bone bone = null;
            bool cachehit = false;
            if (BoneCacheList.TryGetValue(bonename, out bone))
            {
                cachehit = true;
            }
            if (bone == null) bone = skeleton.FindBone(bonename);

            if (!cachehit) BoneCacheList.Add(bonename,bone);

            if (bone == null) return Misc.IdentityMatrix4;
            return ExtractWorldMatrix(bone);
        }

        protected Matrix4 ExtractWorldMatrix(Spine.Bone bone)
        {
            Matrix4 r = Misc.IdentityMatrix4;
            r._11 = bone.A;
            r._12 = bone.C;
            r._21 = bone.B;
            r._22 = bone.D;
            r._41 = (bone.WorldX - skeleton.X) * AntiScaling;
            r._42 = (bone.WorldY - skeleton.Y) * AntiScaling;

            float sx = skeleton.ScaleX < 0.0f ? -1.0f : +1.0f;
            float sy = skeleton.ScaleY < 0.0f ? -1.0f : +1.0f;      // applied YDown
            Misc.m4PushScale(ref r, sx, sy, 1.0f);

            return r;
        }
    }

    public class SpinePhysicsRenderer : SpineRenderer
    {
        public override void ApplyPhysics(double deltatime)
        {
            float inBlendRotate(float r1, float r2, float alpha)
            {
                r2 = r2 - r1;
                r2 -= (16384 - (int)(16384.499999999996 - r2 / 360)) * 360;
                r2 = r1 + r2 * alpha;
                r2 -= (16384 - (int)(16384.499999999996 - r2 / 360)) * 360;

                return r2;
            }
            void inApplyBone(PhysicsGroupSpringBone g, Spine.Skeleton skeleton)
            {
                var alpha = (float)g.Alpha;
                var rs = 1.0 / AntiScaling;

                foreach (var n in g.Nodes)
                {
                    var b = n as PhysicsNodeBone;
                    if (b.UserObject != null)
                    {
                        var sb = b.UserObject as Spine.Bone;
                        switch (g.Mode)
                        {
                            case PhysicsGroupSpringBoneMode.Bone:
                                sb.Rotation = inBlendRotate(sb.Rotation, sb.Rotation + (float)(b.aLocalAngle - b.LocalAngle) * 360.0f, alpha);
                                break;
                            case PhysicsGroupSpringBoneMode.StrBone:
                                sb.Rotation = inBlendRotate(sb.Rotation, sb.Rotation + (float)(b.aLocalAngle - b.LocalAngle) * 360.0f, alpha);
                                sb.X = sb.X + (float)((b.aLocalPosition.x - b.LocalPosition.x) * alpha * rs);
                                sb.Y = sb.Y + (float)((b.aLocalPosition.y - b.LocalPosition.y) * alpha * rs);
                                break;
                            case PhysicsGroupSpringBoneMode.Point:
                            case PhysicsGroupSpringBoneMode.StrPoint:
                                sb.X = sb.X + (float)((b.aLocalPosition.x - b.LocalPosition.x) * alpha * rs);
                                sb.Y = sb.Y + (float)((b.aLocalPosition.y - b.LocalPosition.y) * alpha * rs);
                                break;
                        }
                    }
                }
            }
            void inUpdateMatrix(Spine.Bone bone)
            {
                if (bone == null) return;

                bone.UpdateWorldTransform();

                foreach (var b in bone.Children)
                {
                    inUpdateMatrix(b);
                }
            }
            void inSpring(PhysicsGroupSpringBone g)
            {
                if (g == null) return;

                var m = ExtractWorldMatrix(g.RootUserObject as Spine.Bone);
                Misc.m43PushMatrix(ref m, ref TransMatrix);
                g.RootMatrix = m;
                g.ExternalMatrix = PhysicsExternalMatrix;

                g.Update(deltatime);
                g.Apply();

                inApplyBone(g, skeleton);
                inUpdateMatrix(g.RootBone.UserObject as Spine.Bone);
            }
            void inElement(PhysicsGroupElementBone g)
            {
                if (g == null) return;

                var sb = g.UserObject as Spine.Bone;
                if (sb == null) return;

                g.Position.Value.SetVector(sb.X, sb.Y, 0);
                g.Scale.Value.SetVector(sb.ScaleX, sb.ScaleY, 0);
                g.Rotate.Value.SetVector(0, 0, sb.Rotation / 360.0);   // rotate : need 360deg = 1.0 scale

                g.Update(deltatime);
                g.Apply();

                sb.X = (float)g.Position.aValue.x;
                sb.Y = (float)g.Position.aValue.y;
                sb.ScaleX = (float)g.Scale.aValue.x;
                sb.ScaleY = (float)g.Scale.aValue.y;
                sb.Rotation = (float)(g.Rotate.aValue.z * 360.0);
        }
            // safe check
            if (state == null) return;
            if (skeleton == null) return;

            bool worldupdate = false;
            foreach (var g in PhysicsGroups)
            {
                switch (g.GroupType)
                {
                    case PhysicsGroupType.Spring:
                        inSpring(g as PhysicsGroupSpringBone);
                        break;
                    case PhysicsGroupType.Element:
                        inElement(g as PhysicsGroupElementBone);
                        worldupdate = true;
                        break;
                }
            }
            if (worldupdate) skeleton.UpdateWorldTransform();
        }
        public override void Update(double deltatime, double deltaphysics=-1)
        {
            if (state == null) return;
            if (skeleton == null) return;

            state.Update((float)deltatime);
            PreparePhysics();       // Prepare physics target bones. Set setup-pose.
            state.Apply(skeleton);
            skeleton.UpdateWorldTransform();

            // physics
            if (PhysicsGroups.Count > 0)
            {
                if (deltaphysics < 0) deltaphysics = deltatime;
                ApplyPhysics(deltaphysics);
            }
        }
        public override void AddPhysicsForce(string target, double x, double y, double z)
        {
            // safe check
            if (state == null) return;
            if (skeleton == null) return;

            if (target == "none") return;

            var v = new PhysicsVector3(x, y, z);
            foreach (var g in PhysicsGroups)
            {
                if ((g.AliasName == target) || (target == "all")) g.AddForce(ref v);
            }
        }
        public override void SetupPhysics(List<string> setting)
        {
            // get setting value
            double inGetSettingFloat(Dictionary<string, string> dic, int index, string key, double def_, double min_, double max_)
            {
                string s = "";
                try
                {
                    if (dic.ContainsKey(key)) s = dic[key];
                }
                catch
                {
                    s = "";
                }

                var sl = Misc.DivideString(',', s);
                if (index < sl.Count) s = sl[index];

                var f = Misc.ToFloatDef(s, (float)def_);
                return Math.Min(Math.Max(f, min_), max_);
            }
            string inGetSettingString(Dictionary<string, string> dic, string key, string def_)
            {
                string s = "";
                try
                {
                    if (dic.ContainsKey(key)) s = dic[key];
                }
                catch
                {
                    s = "";
                }
                if (s == "") s = def_;
                return s;
            }
            // spring bone
            void inAddBone(PhysicsGroupSpringBone g, Spine.Bone bone, PhysicsNodeBone parent)
            {
                if (parent != null)
                {
                    var s = bone.Data.Name;
                    if (s.IndexOf("@end") > -1) return;  // terminate code
                    if (s.IndexOf("@") > -1) return;     // when other code exists, safe exit.
                }

                var b = new PhysicsNodeBone(parent);
                b.UserObject = bone;
                b.LocalAngle = bone.Data.Rotation / 360.0;
                b.LocalLength = bone.Data.Length * AntiScaling;
                b.LocalPosition.x = bone.Data.X * AntiScaling;
                b.LocalPosition.y = bone.Data.Y * AntiScaling;
                b.LocalScaleX = bone.Data.ScaleX;
                b.LocalScaleY = bone.Data.ScaleY;
                g.AddBone(b);

                foreach (var c in bone.Children)
                {
                    inAddBone(g, c, b);
                }

                // create terminate bone
                if (bone.Children.Count == 0)
                {
                    parent = b;
                    b = new PhysicsNodeBone(parent);
                    b.UserObject = null;
                    b.LocalAngle = 0;
                    b.LocalLength = 100;
                    b.LocalPosition.x = parent.LocalLength;
                    b.LocalPosition.y = 0;
                    g.AddBone(b);
                }
            }
            void inAddSpring(PhysicsGroupSpringBone g, PhysicsNodeBone bone, PhysicsSpringBone spring)
            {
                foreach (var b in bone.Children)
                {
                    var s = new PhysicsSpringBone(g, spring, bone, b);
                    g.AddSpring(s);
                    inAddSpring(g, b, s);
                }
            }
            void inAddSpringGroup(Dictionary<string, string> dic, string name, string alias, Spine.Bone bone)
            {
                // need has parent
                var root = bone.Parent;
                if (root == null) return;

                if (alias == "") alias = name;

                // enable / disable
                if (inGetSettingString(dic , alias + ".enable", "true") == "false") return;

                var g = new PhysicsGroupSpringBone();
                g.Name = name;
                g.AliasName = alias;
                g.RootUserObject = root;

                var s = inGetSettingString(dic, alias + ".mode", "bone");
                if (s == "strpoint") g.Mode = PhysicsGroupSpringBoneMode.StrPoint;
                if (s == "point") g.Mode = PhysicsGroupSpringBoneMode.Point;
                if (s == "strbone") g.Mode = PhysicsGroupSpringBoneMode.StrBone;
                if (s == "bone") g.Mode = PhysicsGroupSpringBoneMode.Bone;

                g.K = inGetSettingFloat(dic, 0, alias + ".k", g.K, 0.1, 100 * 100);
                g.Damping = inGetSettingFloat(dic, 0, alias + ".damping", g.Damping, 0.1, 100 * 100);
                g.Mass = inGetSettingFloat(dic, 0, alias + ".mass", g.Mass, 0.001, 1000);
                g.LimitVelocity = inGetSettingFloat(dic, 0, alias + ".limitvelocity", g.LimitVelocity, 1, 0xffff);   // 100 = physics frame
                g.Gravity = inGetSettingFloat(dic, 0, alias + ".gravity", g.Gravity, 0, 10000 * 100);
                g.GravityVector.x = inGetSettingFloat(dic, 0, alias + ".gravityvector", g.GravityVector.x, -1000, +1000);
                g.GravityVector.y = inGetSettingFloat(dic, 1, alias + ".gravityvector", g.GravityVector.y, -1000, +1000);
                g.GravityVector.z = inGetSettingFloat(dic, 2, alias + ".gravityvector", g.GravityVector.z, -1000, +1000);
                g.GravityVector.Normalize();
                g.LimitAngleMin = inGetSettingFloat(dic, 0, alias + ".limitangle", g.LimitAngleMin, -0xff, 0);
                g.LimitAngleMax = inGetSettingFloat(dic, 1, alias + ".limitangle", g.LimitAngleMax, 0, +0xff);
                g.LimitMoveRatio = inGetSettingFloat(dic, 0, alias + ".limitmove", g.LimitMoveRatio, 0, +0xff);

                g.ForceRatio = inGetSettingFloat(dic, 0, alias + ".force", g.ForceRatio, -10.0, 10.0);
                g.Alpha = inGetSettingFloat(dic, 0, alias + ".alpha", g.Alpha, 0.0, 1.0);
                g.ForceWiggle = inGetSettingFloat(dic, 0, alias + ".wiggle", g.ForceWiggle, 0.0, 1.0);

                // private
                g.ForceRotateRatio = inGetSettingFloat(dic, 0, alias + ".forcerotate", g.ForceRotateRatio, 0.0, 10.0);
      

                inAddBone(g, bone, null);
                inAddSpring(g, g.RootBone, null);

                g.RootMatrix = ExtractWorldMatrix(root);
                g.SetupNatural();

                PhysicsGroups.Add(g);
            }
            int inGetMask(string s)
            {
                if (s == "none") return 0;

                var m = 0;
                if (s.IndexOf('x', 0) > 0) m = m + 1;
                if (s.IndexOf('y', 0) > 0) m = m + 2;
                if (s.IndexOf('z', 0) > 0) m = m + 4;
                return m;
            }
            void inAddElementGroup(Dictionary<string, string> dic, string name, string alias, Spine.Bone bone)
            {
                if (alias == "") alias = name;

                // enable / disable
                if (inGetSettingString(dic, alias + ".enable", "true") == "false") return;

                var g = new PhysicsGroupElementBone();
                g.Name = name;
                g.AliasName = alias;
                g.UserObject = bone;

                g.Position.Mask = inGetMask(inGetSettingString(dic, alias + ".maskpos", "xyz"));
                g.Scale.Mask = inGetMask(inGetSettingString(dic, alias + ".maskscale", "xyz"));
                g.Rotate.Mask = inGetMask(inGetSettingString(dic, alias + ".maskrot", "xyz"));

                g.K = inGetSettingFloat(dic, 0, alias + ".k", g.K, 0.1, 100 * 100);
                g.Damping = inGetSettingFloat(dic, 0, alias + ".damping", g.Damping, 0.1, 100 * 100);
                g.Mass = inGetSettingFloat(dic, 0, alias + ".mass", g.Mass, 0.001, 1000);
                g.LimitVelocity = inGetSettingFloat(dic, 0, alias + ".limitvelocity", g.LimitVelocity, 1,0xffff);   // 100 = physics frame
                var a = (float)inGetSettingFloat(dic, 0, alias + ".alpha", 1.0, 0.0, +2.0);
                g.Position.Alpha = a;
                g.Scale.Alpha = a;
                g.Rotate.Alpha = a;

                PhysicsGroups.Add(g);
            }

            // do not forget to clear PhysicsGroups
            base.SetupPhysics(setting);

            if (state == null) return;
            if (skeleton == null) return;

            var settingdic = Misc.ConvertKeyValue(setting);

            skeleton.UpdateWorldTransform();

            foreach(var b in skeleton.Bones)
            {
                var s = b.Data.Name;

                var idx = s.IndexOf("@");
                if (idx == -1) continue;

                // name @ spring,aliasname
                var op = "";
                var alias = "";
                var name = s.Substring(0,idx);
                var sl = Misc.DivideString(',', s.Substring(idx + 1));

                if (sl.Count > 0) op = sl[0];
                if (sl.Count > 1) alias = sl[1];

                if (op == "spring") inAddSpringGroup(settingdic, name, alias, b);
                if (op == "element") inAddElementGroup(settingdic, name, alias, b);
            }
        }
        public override void PreparePhysics()
        {
            void inSpring(PhysicsGroupSpringBone g, Spine.Skeleton skeleton)
            {
                if (g == null) return;
                foreach (var n in g.Nodes)
                {
                    var b = n as PhysicsNodeBone;
                    if (b.UserObject != null)
                    {
                        var sb = b.UserObject as Spine.Bone;
                        switch (g.Mode)
                        {
                            case PhysicsGroupSpringBoneMode.Bone:
                                sb.Rotation = sb.Data.Rotation;
                                break;
                            case PhysicsGroupSpringBoneMode.StrBone:
                                sb.Rotation = sb.Data.Rotation;
                                sb.X = sb.Data.X;
                                sb.Y = sb.Data.Y;
                                break;
                            case PhysicsGroupSpringBoneMode.Point:
                            case PhysicsGroupSpringBoneMode.StrPoint:
                                sb.X = sb.Data.X;
                                sb.Y = sb.Data.Y;
                                break;
                        }
                    }
                }
            }
            void inElement(PhysicsGroupElementBone g, Spine.Skeleton skeleton)
            {
                if (g == null) return;
                var sb = g.UserObject as Spine.Bone;
                if (sb == null) return;

                sb.Rotation = sb.Data.Rotation;
                sb.X = sb.Data.X;
                sb.Y = sb.Data.Y;
                sb.ScaleX = sb.Data.ScaleX;
                sb.ScaleY = sb.Data.ScaleY;
            }
            // safe check
            if (state == null) return;
            if (skeleton == null) return;

            foreach (var g in PhysicsGroups)
            {
                switch (g.GroupType)
                {
                    case PhysicsGroupType.Spring:
                        inSpring(g as PhysicsGroupSpringBone, skeleton);
                        break;
                    case PhysicsGroupType.Element:
                        inElement(g as PhysicsGroupElementBone, skeleton);
                        break;
                }
            }
        }
        public override void AppeasePhysics()
        {
            if (state == null) return;
            if (skeleton == null) return;

            foreach (var g in PhysicsGroups)
            {
                g.Appease();
            }
        }
    }
}